import re
from functools import wraps
from typing import Any, overload, List, Union, Callable, Tuple


def quote(token: Any) -> str:
    """ Escapes a token for command line."""
    token = ensure_text(token)
    if re.search(r'[ ()[\];]', token):
        return '"%s"' % token.replace('\"', '\\"')
    return token


@overload
def ensure_binary(x: List[Any]) -> List[bytes]:
    ...


@overload
def ensure_binary(x: Union[int, float, str, bytes, None]) -> bytes:
    ...


@overload
def ensure_binary(x: Callable[..., Tuple[Any]]
                  ) -> Callable[..., Tuple[bytes]]:
    ...


@overload
def ensure_binary(x: Callable[..., List[Any]]
                  ) -> Callable[..., List[bytes]]:
    ...


@overload
def ensure_binary(x: Callable[..., Union[int, float, str, bytes, None]]
                  ) -> Callable[..., bytes]:
    ...


def ensure_binary(x: Any) -> Any:
    """ Recursively ensures that all values except collections are bytes.

    * tuples and lists are encoded recursively
    * strings are encoded with utf-8
    * bytes are left intact
    * functions are decorated with ensuring binary results
    * all other types are converted to string and encoded
    """
    if isinstance(x, tuple):
        return tuple(ensure_binary(y) for y in x)
    if isinstance(x, list):
        return list(ensure_binary(y) for y in x)
    if isinstance(x, str):
        return x.encode("utf-8")
    if callable(x):
        @wraps(x)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            return ensure_binary(x(*args, **kwargs))

        return wrapper
    if not isinstance(x, bytes):
        return str(x).encode("utf-8")
    return x


@overload
def ensure_text(x: Tuple[Any, ...]) -> Tuple[str, ...]:
    ...


@overload
def ensure_text(x: List[Any]) -> List[str]:
    ...


@overload
def ensure_text(x: Union[int, float, str, bytes, None]) -> str:
    ...


def ensure_text(x: Any) -> Any:
    """ Recursively ensures that all values except collections are strings."""
    if isinstance(x, tuple):
        return tuple(ensure_text(y) for y in x)
    if isinstance(x, list):
        return list(ensure_text(y) for y in x)
    if isinstance(x, bytes):
        return x.decode("utf-8")
    return str(x)



import collections

from fffw.graph.meta import StreamType
from fffw.graph import base
from fffw.encoding import inputs, outputs

__all__ = [
    'FilterComplex'
]


class FilterComplex:
    """ ffmpeg filter graph wrapper."""

    def __init__(self, input_list: inputs.InputList,
                 output_list: outputs.OutputList):
        """
        :param input_list: list of input files, containing video and audio
        streams.
        :param output_list: list of output files, with codecs defined.
        """
        self.__input_list = input_list
        self.__output_list = output_list

    def get_free_source(self, kind: StreamType) -> base.Source:
        """
        :param kind: stream type
        :return: first stream of this kind not connected to filter graph
        """
        for stream in self.__input_list.streams:
            if stream.kind != kind or stream.connected:
                continue
            return stream
        else:
            raise RuntimeError("No free streams")

    def render(self, partial: bool = False) -> str:
        """
        Returns filter_graph description in corresponding ffmpeg param syntax.
        """
        result = []
        with base.Namer():
            # Initialize namer context to track unique edge identifiers.
            # In name generation there is no access to namer, so it is accessed
            # via Namer singleton's method. Within context it is guaranteed that
            # same edges will receive same names and different edges will
            # receive unique names. This includes idempotent results for
            # subsequent render() calls for outer Namer context.
            for src in self.__input_list.streams:
                result.extend(src.render(partial=partial))

        # There are no visit checks in recurse graph traversing, so remove
        # duplicates respecting order of appearance.
        return ';'.join(collections.OrderedDict.fromkeys(result))

    def __str__(self) -> str:
        return self.render()



from dataclasses import dataclass
from typing import Optional, List, Tuple, cast, Iterable, Union, Any

from fffw.encoding import filters, outputs
from fffw.graph import base
from fffw.graph.meta import *
from fffw.wrapper import BaseWrapper, param

__all__ = [
    'Input',
    'InputList',
    'Stream',
    'input_file',
]


class Stream(base.Source):
    """ Video or audio stream in input file."""
    source = cast("Input", base.Once('source'))
    """ Source file that contains current stream."""
    index = cast(int, base.Once('index'))
    """ Index of current stream in source file."""

    def __init__(self, kind: StreamType, meta: Optional[Meta] = None):
        """
        :param kind: stream kind, video or audio
        :param meta: stream metadata
        """
        super().__init__(kind=kind, meta=meta)

    @property
    def name(self) -> str:
        if self.index == 0:
            return f'{self.source.index}:{self._kind.value}'
        return f'{self.source.index}:{self._kind.value}:{self.index}'

    def split(self, count: int = 1) -> List[filters.Filter]:
        """
        Splits input stream to reuse it as input node for multiple output nodes.

        >>> from fffw.graph import meta
        >>> stream = Stream(meta.VIDEO)
        >>> s1, s2 = stream.split(2)
        >>> s1 | filters.Scale(1280, 720)
        >>> s2 | filters.Scale(640, 360)
        """
        split = filters.Split(self.kind, output_count=count)
        self.connect_dest(split)
        return [split] * count

    def connect_input(self, source: str) -> None:
        """
        Marks current stream metadata that it belongs to some input file.

        :param source: source filename
        """
        if self.meta is None:
            return
        if self.meta.streams:
            return
        self.meta.streams = [source]
        for scene in self.meta.scenes:
            scene.stream = source


def default_streams() -> Tuple[Stream, ...]:
    """
    Generates default streams definition for input file

    :returns: a tuple with one video and one audio stream.
    """
    return Stream(VIDEO), Stream(AUDIO)


@dataclass
class Input(BaseWrapper):
    # noinspection PyUnresolvedReferences
    """
    Input command line params generator for FFMPEG.

    :arg fast_seek: seek input file over key frames
    :arg input_file: input file name
    :arg slow_seek: perform whole file decoding and output frames only
        from offset to end of file.
    :arg duration: stop decoding frames after an interval
    """
    index = cast(int, base.Once("index"))
    """ Internal ffmpeg source file index."""
    streams: Tuple[Stream, ...] = param(default=default_streams, skip=True)
    """ List of audio and video streams for input file."""

    hardware: str = param(name='hwaccel')
    device: str = param(name='hwaccel_device')
    output_format: str = param(name='hwaccel_output_format')
    fast_seek: Union[TS, float, int] = param(name='ss')
    duration: Union[TS, float, int] = param(name='t')
    input_file: str = param(name='i')
    slow_seek: Union[TS, float, int] = param(name='ss')

    def __post_init__(self) -> None:
        """
        Enumerate streams in input file and freeze instance.
        """
        self.__link_streams_to_input()
        super().__post_init__()

    def __or__(self, other: filters.Filter) -> filters.Filter:
        """
        Connect first available stream to a filter.
        """
        if not isinstance(other, filters.Filter):
            return NotImplemented
        return self.get_stream(other.kind) | other

    def __gt__(self, other: outputs.Codec) -> outputs.Codec:
        """
        Connect first available stream to a codec.
        """
        if not isinstance(other, outputs.Codec):
            return NotImplemented
        return self.get_stream(other.kind) > other

    def __link_streams_to_input(self) -> None:
        """
        Add a link to self to input streams and enumerate streams to get
        proper stream index for input.
        """
        video_streams = 0
        audio_streams = 0
        if self.streams is None:
            raise RuntimeError("Streams not initialized")
        for stream in self.streams:
            if stream.kind == VIDEO:
                meta: Optional[VideoMeta] = getattr(stream, 'meta', None)
                if self.hardware and self.device and meta:
                    meta.device = Device(hardware=self.hardware,
                                         name=self.device)
                stream.index = video_streams
                video_streams += 1
            elif stream.kind == AUDIO:
                stream.index = audio_streams
                audio_streams += 1
            else:
                raise ValueError(stream.kind)
            stream.source = self

    @property
    def audio(self) -> Stream:
        return self.get_stream(AUDIO)

    @property
    def video(self) -> Stream:
        return self.get_stream(VIDEO)

    def get_stream(self, kind: StreamType) -> Stream:
        """
        :param kind: desired stream kind
        :return: first available stream of desired kind
        :raises KeyError: if no streams of this kind found.
        """
        for stream in self.streams:
            if stream.kind == kind:
                return stream
        raise KeyError(kind)


def input_file(filename: str, *streams: Stream, **kwargs: Any) -> Input:
    kwargs['input_file'] = filename
    if streams:
        for stream in streams:
            stream.connect_input(filename)
        kwargs['streams'] = streams
    return Input(**kwargs)


class InputList(list):
    """ List of inputs in FFMPEG."""

    def __init__(self, sources: Iterable[Input] = ()) -> None:
        """
        :param sources: list of input files
        """
        super().__init__()
        self.extend(sources)

    @property
    def streams(self) -> List[Stream]:
        result: List[Stream] = []
        for source in self:
            if source.streams is None:
                raise RuntimeError("Source streams not initialized")
            result.extend(source.streams)
        return result

    def append(self, source: Input) -> None:
        """
        Adds new source file to input list.

        :param source: input file
        """
        source.index = len(self)
        super().append(source)

    def extend(self, sources: Iterable[Input]) -> None:
        """
        Adds multiple source files to input list.

        :param sources: list of input files
        """
        for i, source in enumerate(sources, start=len(self)):
            source.index = i
        super().extend(sources)

    def get_args(self) -> List[bytes]:
        result: List[bytes] = []
        for source in self:
            result.extend(source.get_args())
        return result



from dataclasses import field, dataclass, Field, fields
from typing import Any, Optional, Tuple, cast, List, Callable


def param(default: Any = None, name: Optional[str] = None,
          stream_suffix: bool = False, init: bool = True, skip: bool = False,
          render: Callable[[Any], Any] = None
          ) -> Any:
    """
    Command line and filter parameters constructor.

    :param default: default value for parameter, or a callable (see dataclass
        field default_factory)
    :param name: command line argument name
    :param stream_suffix: flag to add stream specifier suffix to parameter name
        (like '-b:v' or '-c:a')
    :param init: add parameter to dataclass constructor.
    :param skip: skip parameter while generating command line args list.
    :param render: a callable used to format output value.
    :return: dataclass field definition with extra metadata.

    >>> from fffw.encoding.filters import VideoFilter
    >>> @dataclass
    ... class Deinterlace(VideoFilter):
    ...     filter = 'yadif'
    ...     mode: int = param(default=0)
    ...
    >>>
    """
    metadata = {
        'name': name,
        'stream_suffix': stream_suffix,
        'skip': skip,
        'render': render
    }
    if callable(default):
        return field(default_factory=default, init=init, metadata=metadata)
    else:
        return field(default=default, init=init, metadata=metadata)


_FROZEN = '__frozen__'


@dataclass
class Params:
    """ Base class for parametrized objects."""
    ALLOWED = cast(Tuple[str], tuple())
    """ 
    List of attributes that are allowed to be set after instance initialization.
    """

    def __post_init__(self) -> None:
        """ Marks class instance as frozen."""
        setattr(self, _FROZEN, True)

    def __setattr__(self, key: str, value: Any) -> None:
        """ If class is frozen, forbids instance attributes modification."""
        frozen = getattr(self, _FROZEN, False)
        allowed = self.ALLOWED
        if frozen and key not in allowed:
            raise RuntimeError("Parameters are frozen")
        object.__setattr__(self, key, value)

    @property
    def _fields(self) -> Tuple[Field, ...]:
        """
        :return: ordered list of dataclass field
        """
        return fields(self)

    def as_pairs(self) -> List[Tuple[Optional[str], Optional[str]]]:
        """
        :return: a list or pairs (key, value), where key is optional ffmpeg
            parameter name and value is parameter value

        * parameters are all dataclass fields without `skip` flag.
        * by default field name is used as ffmpeg parameter name, but this
          can be overridden by `name` metadata field.
        * if `name` is empty, key is omitted.
        * if `stream_suffix` metadata flag is set, `:v` or `:a` modifier is
          appended to key depending of current object stream kind.
        * if value is callable, real value is computed via function call
        * if value is iterable, corresponding parameter is added to a result
          multiple times
        * if value is `True`, parameter is added as a flag without a value
        """
        args = cast(List[Tuple[Optional[str], Optional[str]]], [])
        for f in self._fields:  # type: Field
            key = f.name
            value = getattr(self, key)
            if f.default == value and f.init:
                # if field value has default value and is configurable via
                # __init__, we omit this field
                continue
            if not value:
                # if value is not set, we omit this field
                continue

            meta = f.metadata
            name = meta.get('name')
            stream_suffix = meta.get('stream_suffix')
            skip = meta.get('skip')
            render = meta.get('render')
            if skip:
                # if field metadata is marked as `skip`
                continue
            if name is None:
                # by default field name is used as parameter name
                name = key
            if stream_suffix:
                # append stream suffix (':v' or ':a') to parameter name
                name = f'{name}:{getattr(self, "kind").value}'
            if callable(render):
                value = render(value)

            if callable(value):
                # get lazy parameter value
                value = value()

            if isinstance(value, (list, tuple)):
                # output multiple parameter values (-i file1 -i file2 ...)
                args.extend((name, str(v)) for v in value)
            elif value is True:
                # support flag without a value
                assert name
                args.append((name, None))
            else:
                # output parameter name and value converted to string
                args.append((name, str(value)))
        return args



from dataclasses import dataclass, replace, asdict, field, fields
from typing import Union, List, cast

from fffw.graph import base
from fffw.encoding import mixins
from fffw.graph.meta import Meta, VideoMeta, TS, Scene, VIDEO, AUDIO, AudioMeta
from fffw.graph.meta import StreamType, Device
from fffw.wrapper.params import Params, param

__all__ = [
    'AutoFilter',
    'AudioFilter',
    'VideoFilter',
    'Concat',
    'Format',
    'Overlay',
    'Scale',
    'SetPTS',
    'Split',
    'Trim',
    'Upload',
    'ensure_video',
    'ensure_audio',
]


def ensure_video(meta: Meta, *_: Meta) -> VideoMeta:
    """
    Checks that first passed stream is a video stream

    :returns: first passed stream
    """
    if not isinstance(meta, VideoMeta):
        raise TypeError(meta)
    return meta


def ensure_audio(meta: Meta, *_: Meta) -> AudioMeta:
    """
    Checks that first passed stream is a audio stream

    :returns: first passed stream
    """
    if not isinstance(meta, AudioMeta):
        raise TypeError(meta)
    return meta


@dataclass
class Filter(mixins.StreamValidationMixin, base.Node, Params):
    """
    Base class for ffmpeg filter definitions.

    `VideoFilter` and `AudioFilter` are used to define new filters.
    """
    ALLOWED = ('enabled',)
    """ fields that are allowed to be modified after filter initialization."""

    @property
    def args(self) -> str:
        """ Formats filter args as k=v pairs separated by colon."""
        args = self.as_pairs()
        result = []
        for key, value in args:
            if key and value:
                result.append(f'{key}={value}')
        return ':'.join(result)

    def split(self, count: int = 1) -> List["Filter"]:
        """
        Adds a split filter to "fork" current node and reuse it as input node
        for multiple destinations.

        >>> d = VideoFilter("deint")
        >>> d1, d2 = d.split(count=2)
        >>> d1 | Scale(1280, 720)
        >>> d2 | Scale(480, 360)

        :returns: a list with `count` copies of Split() filter. These are
        multiple references for the same object, so each element is intended
        to be reused only once.
        """
        split = Split(self.kind, output_count=count)
        self.connect_dest(split)
        return [split] * count

    def clone(self, count: int = 1) -> List["Filter"]:
        """
        Creates multiple copies of self to reuse it as output node for multiple
        sources.

        Any connected input node is being split and connected to a corresponding
        copy of current filter.
        """
        if count == 1:
            return [self]
        result = []
        for _ in range(count):
            result.append(self._clone())

        for i, edge in enumerate(self.inputs):
            if edge is None:
                continue
            # reconnecting incoming edge to split filter
            split = Split(self.kind, output_count=count)
            edge.reconnect(split)
            for dst in result:
                split | dst

        return result

    def _clone(self) -> "Filter":
        """
        Creates a copy of current filter with same parameters.

        Inputs and outputs are not copied.
        """
        skip = [f.name for f in fields(self) if not f.init]
        kwargs = {k: v for k, v in asdict(self).items() if k not in skip}
        # noinspection PyArgumentList
        return type(self)(**kwargs)  # type: ignore


class VideoFilter(Filter):
    """
    Base class for video filters.

    >>> @dataclass
    ... class Deinterlace(VideoFilter):
    ...     filter = 'yadif'
    ...     mode: int = param(default=0)
    ...
    >>>
    """
    kind = VIDEO


class AudioFilter(Filter):
    """
    Base class for audio filters.

    >>> @dataclass
    ... class Volume(AudioFilter):
    ...     filter = 'volume'
    ...     volume: float = param(default=1.0)
    ...
    >>>
    """
    kind = AUDIO


@dataclass
class AutoFilter(Filter):
    """
    Base class for stream kind autodetect.
    """

    kind: StreamType = field(metadata={'skip': True})
    """ 
    Stream kind used to generate filter name. Required. Not used as filter 
    parameter.
    """

    # `field` is used here to tell MyPy that there is no default for `kind`
    # because `default=MISSING` is valuable for MyPY.

    def __post_init__(self) -> None:
        """ Adds audio prefix to filter name for audio filters."""
        if self.kind == AUDIO:
            self.filter = f'a{self.filter}'
        super().__post_init__()


@dataclass
class Scale(VideoFilter):
    # noinspection PyUnresolvedReferences
    """
    Video scaling filter.

    :arg width: resulting video width
    :arg height: resulting video height
    """
    filter = "scale"
    hardware = None  # cpu only

    width: int = param(name='w')
    height: int = param(name='h')

    def transform(self, *metadata: Meta) -> Meta:
        meta = ensure_video(*metadata)
        par = meta.dar / (self.width / self.height)
        return replace(meta, width=self.width, height=self.height, par=par)


@dataclass
class Split(AutoFilter):
    # noinspection PyUnresolvedReferences
    """
    Audio or video split filter.

    Splits audio or video stream to multiple output streams (2 by default).
    Unlike ffmpeg `split` filter this one does not allow to pass multiple
    inputs.

    :arg kind: stream type.
    :arg output_count: number of output streams.
    """
    filter = 'split'

    output_count: int = 2

    def __post_init__(self) -> None:
        """
        Disables filter if `output_count` equals 1 (no real split is preformed).
        """
        self.enabled = self.output_count > 1
        super().__post_init__()

    @property
    def args(self) -> str:
        """
        :returns: split/asplit filter parameters
        """
        if self.output_count == 2:
            return ''
        return str(self.output_count)


@dataclass
class Trim(AutoFilter):
    # noinspection PyUnresolvedReferences
    """
    Cuts the input so that the output contains one continuous subpart
    of the input.

    :arg kind: stream kind (for proper concatenated streams definitions).
    :arg start: start time of trimmed output
    :arg end: end time of trimmed output
    """
    filter = 'trim'

    start: Union[int, float, str, TS]
    end: Union[int, float, str, TS]

    def __post_init__(self) -> None:
        if not isinstance(self.start, (TS, type(None))):
            self.start = TS(self.start)
        if not isinstance(self.end, (TS, type(None))):
            self.end = TS(self.end)
        super().__post_init__()

    def transform(self, *metadata: Meta) -> Meta:
        """
        Computes metadata for trimmed stream.

        :param metadata: single incoming stream metadata.
        :returns: metadata with initial start (this is fixed with SetPTS) and
            duration set to trim end. Scenes list is intersected with trim
            interval, scene borders are aligned to trim borders.
        """
        meta = metadata[0]
        scenes = []
        streams: List[str] = []
        for scene in meta.scenes:
            if scene.stream and (not streams or streams[0] != scene.stream):
                # Adding an input stream without contiguous duplicates.
                streams.append(scene.stream)

            # intersect scene with trim interval
            start = cast(TS, max(self.start, scene.start))
            end = cast(TS, min(self.end, scene.end))

            if start < end:
                # If intersection is not empty, add intersection to resulting
                # scenes list.
                # This will allow to detect buffering when multiple scenes are
                # reordered in same file: input[3:4] + input[1:2]
                scenes.append(Scene(stream=scene.stream, start=start,
                                    duration=end - start))

        kwargs = {
            'start': self.start,
            'duration': self.end,
            'scenes': scenes,
            'streams': streams
        }
        interval = cast(TS, self.end) - cast(TS, self.start)
        if isinstance(meta, AudioMeta):
            kwargs['samples'] = round(meta.sampling_rate * interval)
        if isinstance(meta, VideoMeta):
            kwargs['frames'] = round(meta.frame_rate * interval)
        return replace(meta, **kwargs)


@dataclass
class SetPTS(AutoFilter):
    """
    Change the PTS (presentation timestamp) of the input frames.

    $  ffmpeg -y -i source.mp4 \
    -vf trim=start=3:end=4,setpts=PTS-STARTPTS -an test.mp4

    Supported cases for metadata handling:

    * "PTS-STARTPTS" - resets stream start to zero.
    """
    RESET_PTS = 'PTS-STARTPTS'
    filter = 'setpts'

    expr: str = param(default=RESET_PTS)

    def transform(self, *metadata: Meta) -> Meta:
        meta = metadata[0]
        expr = self.expr.replace(' ', '')
        if expr == self.RESET_PTS:
            duration = meta.duration - meta.start
            return replace(meta, start=TS(0), duration=duration)
        raise NotImplementedError()

    @property
    def args(self) -> str:
        return self.expr


@dataclass
class Concat(Filter):
    # noinspection PyUnresolvedReferences
    """
    Concatenates multiple input stream to a single one.

    :arg kind: stream kind (for proper concatenated streams definitions).
    :arg input_count: number of input streams.
    """
    filter = 'concat'

    kind: StreamType
    input_count: int = 2

    def __post_init__(self) -> None:
        """
        Disables filter if input_count is 1 (no actual concatenation is
        performed).
        """
        self.enabled = self.input_count > 1
        super().__post_init__()

    @property
    def args(self) -> str:
        if self.kind == VIDEO:
            if self.input_count == 2:
                return ''
            return 'n=%s' % self.input_count
        return 'v=0:a=1:n=%s' % self.input_count

    def transform(self, *metadata: Meta) -> Meta:
        """
        Compute metadata for concatenated streams.

        :param metadata: concatenated streams metadata
        :returns: Metadata for resulting stream with duration set to a sum of
            stream durations. Scenes and streams are also concatenated.
        """
        duration = TS(0)
        scenes = []
        streams: List[str] = []
        frames: int = 0
        for meta in metadata:
            duration += meta.duration
            scenes.extend(meta.scenes)
            for stream in meta.streams:
                if not streams or streams[-1] != stream:
                    # Add all streams for each concatenated metadata and remove
                    # contiguous duplicates.
                    streams.append(stream)
            if isinstance(meta, VideoMeta):
                frames += meta.frames
        kwargs = dict(duration=duration, scenes=scenes, streams=streams)
        meta = metadata[0]
        if isinstance(meta, AudioMeta):
            # Recompute samples and sampling rate: sampling rate from first
            # input, samples count corresponds duration.
            kwargs['samples'] = round(meta.sampling_rate * duration)
        if isinstance(meta, VideoMeta):
            # Sum frames count from all input streams
            kwargs['frames'] = frames
        return replace(metadata[0], **kwargs)


@dataclass
class Overlay(VideoFilter):
    # noinspection PyUnresolvedReferences
    """
    Combines two video streams one on top of another.

    :arg x: horizontal position of top image in bottom one.
    :arg y: vertical position of top image in bottom one.
    """
    input_count = 2
    filter = "overlay"

    x: int
    y: int


@dataclass
class Format(VideoFilter):
    """
    Converts pixel format of video stream
    """
    filter = 'format'
    format: str = param(name='pix_fmts')


@dataclass
class Upload(VideoFilter):
    """
    Uploads a stream to a hardware device
    """
    filter = 'hwupload'
    extra_hw_frames: int = param(default=64, init=False)
    device: Device = param(skip=True)

    def transform(self, *metadata: Meta) -> VideoMeta:
        """ Marks a stream as uploaded to a device."""
        meta = ensure_video(*metadata)
        return replace(meta, device=self.device)



import asyncio
import io
import subprocess
from asyncio.subprocess import Process
from dataclasses import dataclass
from logging import getLogger
from types import TracebackType
from typing import Tuple, List, Any, Optional, cast, Callable, Union, TextIO
from typing import Type, AsyncIterator

from fffw.wrapper.helpers import quote, ensure_binary, ensure_text
from fffw.wrapper.params import Params


class UniversalLineReader:
    """
    Reads bytes from asyncio.StreamReader and splits it to lines with either
    CR or LF, or even CRLF.

    https://docs.python.org/3/glossary.html#term-universal-newlines

    >>> # noinspection PyUnresolvedReferences
    ... line_iter = UniversalLineReader(process.stderr)
    >>> async for line in line_iter:
    ...    print(line)

    """

    def __init__(self,
                 reader: asyncio.StreamReader,
                 bufsize: int = 10 * io.DEFAULT_BUFFER_SIZE,
                 blocksize: int = io.DEFAULT_BUFFER_SIZE,
                 encoding: str = 'utf8') -> None:
        """
        :param reader: asynchronous stream reader, i.e. stdout/stderr of
            asyncio Process instance.
        :param bufsize: max buffer size
        :param blocksize: read block size
        :param encoding: text encoding
        """
        self.blocksize = blocksize
        self.reader = reader
        self.bufsize = bufsize
        self.buffer = b''
        self.encoding = encoding
        self.at_eof = False

    def __aiter__(self) -> AsyncIterator[str]:
        return self.readlines()

    async def readlines(self) -> AsyncIterator[str]:
        while not self.at_eof:
            # StreamReader supports only LF line separator. This leads to buffer
            # overrun when it contains only CR-terminated lines. Thus, we read
            # blocks manually and then split it to lines with universal line
            # separator.
            block = await self.reader.read(self.blocksize)
            # empty read means that stream is closed
            self.at_eof = len(block) == 0

            # checking max buffer size
            buffered = len(self.buffer) + len(block)
            if buffered > self.bufsize:
                raise asyncio.LimitOverrunError("buffer overrun", buffered)

            self.buffer += block
            if self.buffer:
                # Split buffer to line with any of CR, LF of CRLF separators
                # Last line is buffered to handle the case when CRLF sequence is
                # being split to subsequent reads.
                [*lines, self.buffer] = self.buffer.splitlines(keepends=True)
                for line in lines:
                    yield line.decode(self.encoding)
        # We always leave one non-empty line above, but stream may be empty. In
        # this case we don't want to yield empty line.
        if self.buffer:
            yield self.buffer.decode(self.encoding)


class Runner:
    """ Wrapper for Popen process for non-blocking streams handling."""

    def __init__(self, command: Union[str, bytes], *args: Any,
                 stdin: Union[None, str, TextIO] = None,
                 stdout: Optional[Callable[[str], str]] = None,
                 stderr: Optional[Callable[[str], str]] = None,
                 timeout: Union[int, float, None] = None) -> None:
        """
        :param command: executable file name
        :param args: executable arguments
        :param stdin: input stream or content
        :param stdout: output stream line handler
        :param stderr: error stream line handler
        :param timeout: process execution timeout
        """
        self.command = ensure_binary(command)
        self.args = ensure_binary(list(args))
        if isinstance(stdin, str):
            stdin = io.StringIO(stdin)
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.timeout = timeout
        self.output = io.StringIO()
        self.errors = io.StringIO()

    async def start_process(self) -> Process:
        """ Starts child process."""
        return await asyncio.create_subprocess_exec(
            self.command, *self.args,
            stdin=self.stdin and subprocess.PIPE,
            stdout=self.stdout and subprocess.PIPE,
            stderr=self.stderr and subprocess.PIPE,
        )

    async def __aenter__(self) -> Process:
        """ Starts child process and initialize I/O handling."""
        self.proc = await self.start_process()
        return self.proc

    async def __aexit__(self,
                        exc_type: Optional[Type[BaseException]],
                        exc_val: Optional[BaseException],
                        exc_tb: TracebackType) -> None:
        await self.proc.wait()

    def __call__(self) -> Tuple[int, str, str]:
        return asyncio.get_event_loop().run_until_complete(self.run())

    @staticmethod
    async def read(reader: Optional[asyncio.StreamReader],
                   callback: Optional[Callable[[str], str]],
                   output: io.StringIO) -> None:
        """
        Handles read from stdout/stderr.

        Reads lines from stream and feeds it to callback. Values, filtered by
        callback, are written to output buffer.

        :param reader: Process.stdout or Process.stderr instance
        :param callback: callback for handling read lines
        :param output: output biffer
        """
        if callback is None or reader is None:
            return
        async for line in UniversalLineReader(reader):
            data = callback(line)
            if data:
                output.write(data)

    @staticmethod
    async def write(writer: Optional[asyncio.StreamWriter],
                    stream: Optional[TextIO]) -> None:
        """
        Handle write to stdin.

        :param writer: Process.stdin instance
        :param stream: stream to read lines from
        """
        if stream is None or writer is None:
            return
        while True:
            line = stream.readline()
            if not line:
                break
            writer.write(line.encode())
            try:
                await writer.drain()
            except (ConnectionResetError, BrokenPipeError):
                # inspired by Process.communicate
                return
        writer.close()

    async def run(self) -> Tuple[int, str, str]:
        async with self as p:
            try:
                await asyncio.wait_for(
                    asyncio.shield(asyncio.gather(
                        # write input data to stdin
                        self.write(p.stdin, self.stdin),
                        # read output from stdout
                        self.read(p.stdout, self.stdout, self.output),
                        # read output from stderr
                        self.read(p.stderr, self.stderr, self.errors),
                        # wait for process termination
                        p.wait(),
                    )),
                    timeout=self.timeout)
            except asyncio.TimeoutError:
                self.proc.kill()
        return (cast(int, self.proc.returncode),
                self.output.getvalue(),
                self.errors.getvalue())


# noinspection PyTypeChecker
class CommandMixin:
    command: str
    key_prefix: str = '-'
    key_suffix: str = ' '
    stdout: bool = True
    stderr: bool = True
    timeout: Optional[float] = None
    runner_class = Runner


@dataclass
class BaseWrapper(CommandMixin, Params):
    """
    Base class for generating command line arguments from params.

    Values meanings:
    * True: flag presence
    * False/None: flag absence
    * List/Tuple: param name is repeated multiple times with values
    * Callable: function call result is added to result
    * All other: param name and value are added to result
    """

    def __post_init__(self) -> None:
        cls = self.__class__
        self.logger = getLogger("%s.%s" % (cls.__module__, cls.__name__))

    @ensure_binary
    def get_args(self) -> List[Any]:
        args: List[str] = []
        for key, value in self.as_pairs():
            tokens = []
            if key:
                tokens.append(f'{self.key_prefix}{key}')
            if value:
                tokens.append(value)
            if tokens:
                if self.key_suffix == ' ':
                    args.extend(tokens)
                else:
                    args.append(self.key_suffix.join(tokens))
        return args

    def get_cmd(self) -> str:
        command_line = [self.command] + ensure_text(self.get_args())
        return ' '.join(map(quote, command_line))

    def handle_stderr(self, line: str) -> str:
        if line != '':
            self.logger.debug(line.strip())
        return line

    def handle_stdout(self, line: str) -> str:
        if line != '':
            self.logger.debug(line.strip())
        return line

    def run(self,
            stdin: Union[None, str, TextIO] = None,
            timeout: Optional[float] = None) -> Tuple[int, str, str]:
        args = self.get_args()
        self.logger.debug(self.get_cmd())
        runner = self.runner(
            self.command, *args,
            stdin=stdin,
            stdout=self.handle_stdout if self.stdout else None,
            stderr=self.handle_stderr if self.stderr else None,
            timeout=timeout
        )
        return runner()

    def runner(self,
               command: Union[str, bytes],
               *args: Any,
               stdin: Union[None, str, TextIO] = None,
               stdout: Optional[Callable[[str], str]] = None,
               stderr: Optional[Callable[[str], str]] = None,
               timeout: Union[int, float, None] = None) -> Runner:
        """ Initializer runner instance."""
        return self.runner_class(
            command,
            *args,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            timeout=timeout)



from math import floor, ceil
from typing import Union, Tuple


def xround(val: float, div: int, how: Union[None, str] = None,
           quality: int = 5) -> int:
    """ Number rounding with steroids.

    :param val: float, a number to round
    :param div: int, accuracy. Resulting value will be dividable to this.
    :param how: (None|'ceil'|'floor'), round mode
        None - classic rounding, up or down
        ceil - round up
        floor - round down
    :param quality:
        number of digits after comma used to truncate numbers like
        503.9999999999999994 before main rounding.
    """
    val = round(val, quality)
    if how == 'floor':
        return int(floor(val / float(div))) * div
    if how == 'ceil':
        return int(ceil(val / float(div))) * div
    return (int(val + div / 2.0) // div) * div


class Scaler:
    """ Image and video dimensions transformation helpers."""

    def __init__(self,
                 source_size: Tuple[int, int],
                 par: float = 1.0,
                 rotation: int = 0,
                 accuracy: int = 2):
        self.rotation = rotation in (90, 270)
        w, h = source_size
        self.source_size = (h, w) if self.rotation else (w, h)
        self.accuracy = accuracy
        self.par = par

    def _clone(self) -> "Scaler":
        return Scaler(self.source_size, par=self.par, rotation=self.rotation,
                      accuracy=self.accuracy)

    def crop(self, left: int, top: int, width: int, height: int) -> "Scaler":
        """ Returns new scaler with cut dimensions.
        """
        s = self._clone()
        width = min(self.source_size[0] - left, width)
        height = min(self.source_size[1] - top, height)
        s.source_size = (width, height)
        return s

    def rotate(self, rotation: int = 90) -> "Scaler":
        """ Retuns new scaler with 90-degree rotation."""
        s = self._clone()
        if rotation in (0, 180):
            # width and height are swapped or not
            return s
        height, width = self.source_size
        s.source_size = (width, height)
        s.rotation = not s.rotation
        return s

    @property
    def pixel_size(self) -> Tuple[int, int]:
        """ Source image size in square pixels."""
        sw, sh = self.source_size
        sw = int(sw * self.par)
        return sw, sh

    @property
    def aspect(self) -> float:
        width, height = self.source_size
        if height:
            return round(float(width) / float(height), 3)
        return 0.0

    def scale_fit(self, target_size: Tuple[int, int]) -> "Scaler":
        """
        Scales source image fitting it to
        target_size.

        One or both of resulting dimensions is equal to corresponding target
        dimension, another is less.
        """

        sw, sh = self.pixel_size
        tw, th = target_size
        qsize = sw + sh
        if not qsize:
            raise ValueError("Source size is 0x0")
        scale = min(float(tw) / sw, float(th) / sh)

        return self.scale(scale)

    def scale_crop(self, target_size: Tuple[int, int]
                   ) -> Tuple["Scaler", Tuple[int, int, int, int]]:
        """ Scales source image to target_size with cutting borders.

        One of resulting dimensions is equal to corresponding target dimension,
        another was greated initially, but was cut to target dimension.
        """

        sw, sh = self.pixel_size
        tw, th = target_size
        qsize = sw + sh
        if not qsize:
            raise ValueError("Source size is 0x0")
        scale = max(float(tw) / sw, float(th) / sh)

        s = self.scale(scale)

        # source image part size, cut to target size
        cw, ch = int(tw / scale), int(th / scale)
        left = max(sw - cw, 0) // 2
        top = max(sh - ch, 0) // 2
        return s, (left, top, cw, ch)

    def scale(self, scale: float) -> "Scaler":
        """
        Scales source image with a scale.

        scaled dimension is source dimension multiplied by scale.
        """
        sw, sh = self.pixel_size

        # Usually for horizontal video we round width "up" and height "down".
        # For valid sources it eliminates thin black margins on top/down, for
        # horizontal sources it is not so important.
        (width, height) = (xround(sw * scale, self.accuracy, how='ceil'),
                           xround(sh * scale, self.accuracy, how='floor'))
        s = self._clone()
        s.source_size = (width, height)
        s.par = 1.0
        return s



from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from functools import wraps
from typing import List, Union, Any, Optional, Callable, overload, Tuple, cast
from fffw.types import Literal

from pymediainfo import MediaInfo  # type: ignore

__all__ = [
    'AudioMeta',
    'VideoMeta',
    'StreamType',
    'AUDIO',
    'VIDEO',
    'Device',
    'Meta',
    'Scene',
    'TS',
    'video_meta_data',
    'audio_meta_data',
    'from_media_info',
]


class StreamType(Enum):
    VIDEO = 'v'
    AUDIO = 'a'


VIDEO = StreamType.VIDEO
AUDIO = StreamType.AUDIO

BinaryOp = Callable[[Any, Any], Any]
BinaryTS = Callable[[Any, Any], "TS"]
UnaryOp = Callable[[Any], Any]
UnaryTS = Callable[[Any], "TS"]


@overload
def ts(func: BinaryOp,
       *,
       arg: bool = True,
       res: Literal[True] = True,
       noarg: Literal[False] = False
       ) -> BinaryTS:
    ...


@overload
def ts(func: BinaryOp,
       *,
       arg: bool = True,
       res: Literal[False],
       noarg: Literal[False] = False
       ) -> BinaryOp:
    ...


@overload
def ts(func: UnaryOp,
       *,
       arg: bool = True,
       res: Literal[True] = True,
       noarg: Literal[True]
       ) -> UnaryTS:
    ...


@overload
def ts(func: UnaryOp,
       *,
       arg: bool = True,
       res: Literal[False],
       noarg: Literal[True]
       ) -> UnaryOp:
    ...


def ts(func: Union[BinaryOp, UnaryOp], *,
       arg: bool = True, res: bool = True, noarg: bool = False
       ) -> Union[BinaryOp, UnaryOp]:
    """
    Decorates functions to automatically cast first argument and result to TS.
    """
    if arg and res:
        if noarg:
            @wraps(func)
            def wrapper(self: "TS") -> "TS":
                return TS(cast(UnaryOp, func)(self))  # noqa
        else:
            @wraps(func)
            def wrapper(self: "TS", value: Any) -> "TS":
                if value is None:
                    res = cast(BinaryOp, func)(self, value)  # noqa
                else:
                    res = cast(BinaryOp, func)(self, TS(value))  # noqa
                return TS(res)
    elif arg:
        @wraps(func)
        def wrapper(self: "TS", value: Any) -> Any:
            if value is None:
                return cast(BinaryOp, func)(self, value)  # noqa
            return cast(BinaryOp, func)(self, TS(value))  # noqa
    elif res:
        @wraps(func)
        def wrapper(self: "TS", value: Any) -> "TS":
            return TS(cast(BinaryOp, func)(self, value))  # noqa
    else:
        return func

    return wrapper


class TS(float):
    """
    Timestamp data type.

    Accepts common timestamp formats like '123:45:56.1234'.
    Integer values are parsed as milliseconds.
    """

    def __new__(cls, value: Union[int, float, str, timedelta]) -> "TS":
        """
        :param value: integer duration in milliseconds, float duration in
            seconds or string ffmpeg interval definition (123:59:59.999).
        :returns new timestamp from value.
        """
        if isinstance(value, timedelta):
            value = value.total_seconds()
        elif isinstance(value, int):
            value /= 1000.0
        elif isinstance(value, str):
            if '.' in value:
                value, rest = value.split('.')
                fractional = float(f'0.{rest}')
            else:
                fractional = 0
            seconds = 0
            for part in map(int, value.rsplit(':', 2)):
                seconds *= 60
                seconds += part
            value = seconds + fractional
        return super().__new__(cls, value)  # type: ignore

    __add__ = ts(float.__add__)
    __radd__ = ts(float.__radd__)
    __sub__ = ts(float.__sub__)
    __rsub__ = ts(float.__rsub__)
    __mul__ = ts(float.__mul__, arg=False)
    __rmul__ = ts(float.__rmul__, arg=False)
    __neg__ = ts(float.__neg__, noarg=True)
    __abs__ = ts(float.__abs__, noarg=True)
    __eq__ = ts(float.__eq__, res=False)
    __ne__ = ts(float.__ne__, res=False)
    __gt__ = ts(float.__gt__, res=False)
    __ge__ = ts(float.__ge__, res=False)
    __lt__ = ts(float.__lt__, res=False)
    __le__ = ts(float.__le__, res=False)

    @overload  # type: ignore
    def __floordiv__(self, other: "TS") -> int:
        ...

    @overload  # type: ignore
    def __floordiv__(self, other: int) -> "TS":
        ...

    def __floordiv__(self, other: Union["TS", float, int]) -> Union[int, "TS"]:
        """
        Division behavior from timedelta (rounds to microseconds)

        >>> TS(10.0) // TS(3.0)
        3
        >>> TS(10.0) // 3
        TS(3.333333)
        >>> TS(10.0) // 3.0
        TS(3.333333)
        """
        value = (float(self * 1000000.0) // other) / 1000000.0
        if isinstance(other, TS):
            return int(value)
        return TS(value)

    @overload
    def __truediv__(self, other: "TS") -> float:  # type: ignore
        ...

    @overload
    def __truediv__(self, other: Union[float, int]) -> "TS":  # type: ignore
        ...

    def __truediv__(self, other: Union["TS", float, int]) -> Union[float, "TS"]:
        """
        Division behavior from timedelta

        >>> TS(10.0) / TS(2.125)
        4.705882352941177
        >>> TS(10.0) / 2.125
        TS(4.705882352941177)
        >>> TS(10.0) / 2
        TS(5.0)
        """
        value = super().__truediv__(other)
        if isinstance(other, TS):
            return value
        return TS(value)

    def __divmod__(self, other: float) -> Tuple[int, "TS"]:
        """
        Div/mod behavior from timedelta

        >>> divmod(TS(10.0), TS(2.125))
        (4, TS(1.5))
        """
        div, mod = super().__divmod__(other)
        return int(div), TS(mod)

    def __int__(self) -> int:
        """
        :return: duration in milliseconds.
        """
        return int(float(self * 1000))

    def __str__(self) -> str:
        """
        Removes non-valuable zeros from fractional part.

        :returns: ffmpeg seconds definition (123456.999).
        """
        v = super().__repr__()
        if '.' in v:
            v = v.rstrip('0')
            if v.endswith('.'):
                v += '0'
        return v

    def __repr__(self) -> str:
        return f'TS({super().__repr__()})'

    def total_seconds(self) -> float:
        return float(self)

    @property
    def days(self) -> int:
        return int(float(self / (24 * 3600)))

    @property
    def seconds(self) -> int:
        return int(float(self) % (24 * 3600))

    @property
    def microseconds(self) -> int:
        return int(float(self * 1000000) % 1000000)


@dataclass
class Scene:
    """
    Continuous part of stream used in transcoding graph.
    """
    stream: Optional[str]
    """ Stream identifier."""
    duration: TS
    """ Stream duration."""
    start: TS
    """ First frame/sample timestamp for stream."""

    @property
    def end(self) -> TS:
        return self.start + self.duration


@dataclass
class Meta:
    """
    Stream metadata.

    Describes common stream characteristics like bitrate and duration.
    """
    duration: TS
    """ Resulting stream duration."""
    start: TS
    """ First frame/sample timestamp for resulting stream."""
    bitrate: int
    """ Input stream bitrate in bits per second."""
    scenes: List[Scene]
    """ 
    List of continuous stream fragments (maybe from different files), that need
    to be read to get a result with current metadata.
    """
    streams: List[str]
    """
    List of streams (maybe from different files), that need to be read to get
    a result with current metadata."""

    @property
    def end(self) -> TS:
        """
        :return: Timestamp of last frame resulting stream.
        """
        return self.start + self.duration

    @property
    def kind(self) -> StreamType:
        raise NotImplementedError()


@dataclass
class Device:
    """
    Describes hardware device used for video acceleration
    """
    hardware: str
    name: str


@dataclass
class VideoMeta(Meta):
    """
    Video stream metadata.

    Describes video stream characteristics.
    """
    width: int
    height: int
    par: float
    """ Pixel aspect ratio."""
    dar: float
    """ Display aspect ratio."""
    frame_rate: float
    """ Frames per second."""
    frames: int
    """ Number of frames."""
    device: Optional[Device]
    """ Hardware device asociated with current stream."""

    def __post_init__(self) -> None:
        self.validate()

    @property
    def kind(self) -> StreamType:
        return VIDEO

    def validate(self) -> None:
        if self.height != 0:
            assert abs(self.dar - self.width / self.height * self.par) <= 0.001
        else:
            assert str(self.dar) == 'nan'

        interval = float(self.duration - self.start)
        assert abs(self.frames - interval * self.frame_rate) <= 1


@dataclass
class AudioMeta(Meta):
    """
    Audio stream metadata.

    Describes audio stream characteristics.
    """
    sampling_rate: int
    """ Samples per second."""
    channels: int
    """ Number of audio channels."""
    samples: int
    """ Samples count."""

    def __post_init__(self) -> None:
        self.validate()

    @property
    def kind(self) -> StreamType:
        return AUDIO

    def validate(self) -> None:
        interval = float(self.duration - self.start)
        assert abs(self.samples - interval * self.sampling_rate) <= 1


def audio_meta_data(**kwargs: Any) -> AudioMeta:
    stream = kwargs.get('stream')
    duration = TS(kwargs.get('duration', 0))
    start = TS(kwargs.get('start', 0))
    scene = Scene(
        stream=stream,
        duration=duration,
        start=start,
    )

    return AudioMeta(
        scenes=[scene],
        streams=[stream] if stream else [],
        duration=duration,
        start=start,
        bitrate=int(kwargs.get('bit_rate', 0)),
        channels=int(kwargs.get('channel_s', 0)),
        sampling_rate=int(kwargs.get('sampling_rate', 0)),
        samples=int(kwargs.get('samples_count', 0)),
    )


def video_meta_data(**kwargs: Any) -> VideoMeta:
    duration = TS(kwargs.get('duration', 0))
    width = int(kwargs.get('width', 0))
    height = int(kwargs.get('height', 0))
    par = float(kwargs.get('pixel_aspect_ratio', 1.0))
    try:
        dar = float(kwargs['display_aspect_ratio'])
    except KeyError:
        if height == 0:
            dar = float('nan')
        else:
            dar = width / height * par
    frames = int(kwargs.get('frame_count', 0))
    try:
        frame_rate = float(kwargs['frame_rate'])
    except KeyError:
        if duration.total_seconds() == 0:
            frame_rate = 0
        else:
            frame_rate = frames / duration.total_seconds()

    stream = kwargs.get('stream')
    start = TS(kwargs.get('start', 0))
    scene = Scene(
        stream=stream,
        start=start,
        duration=duration,
    )
    return VideoMeta(
        scenes=[scene],
        streams=[stream] if stream else [],
        duration=duration,
        start=start,
        bitrate=int(kwargs.get('bit_rate', 0)),
        width=width,
        height=height,
        par=par,
        dar=dar,
        frame_rate=frame_rate,
        frames=frames,
        device=None,
    )


def from_media_info(mi: MediaInfo) -> List[Meta]:
    streams: List[Meta] = []
    for track in mi.tracks:
        if track.track_type in ('Video', 'Image'):
            streams.append(video_meta_data(**track.__dict__))
        elif track.track_type == 'Audio':
            streams.append(audio_meta_data(**track.__dict__))
    return streams



from fffw.graph.meta import VIDEO, AUDIO
from fffw.encoding import outputs

__all__ = [
    'AudioCodec',
    'VideoCodec',
]


class VideoCodec(outputs.Codec):
    """
    Base class for describing video codecs.

    See `fffw.encoding.outputs.Codec` for params definition.

    >>> from dataclasses import dataclass
    >>> from fffw.wrapper import param
    >>> @dataclass
    ... class X264(VideoCodec):
    ...     codec = 'libx264'
    ...     gop: int = param(name='g')
    ...
    >>> codec = X264(bitrate=4000000, gop=25)
    >>> copy = VideoCodec('copy')
    """
    kind = VIDEO


class AudioCodec(outputs.Codec):
    """
    Base class for describing audio codecs.

    See :py:class:`fffw.encoding.outputs.Codec` for params definition.

    >>> from dataclasses import dataclass
    >>> from fffw.wrapper import param
    >>> @dataclass
    ... class FdkAAC(AudioCodec):
    ...     codec = 'libfdk_aac'
    ...     rate: int = param(default=48000, name='r')
    ...
    >>> codec = FdkAAC(bitrate=192000, rate=44100)
    >>> copy = AudioCodec('copy')
    """
    kind = AUDIO



from dataclasses import dataclass
from itertools import chain
from typing import List, cast, Optional, Iterable, Any

from fffw.graph.meta import AUDIO, VIDEO, StreamType
from fffw.graph import base
from fffw.wrapper import BaseWrapper, ensure_binary, param
from fffw.encoding import mixins
__all__ = [
    'Codec',
    'Output',
    'OutputList',
    'output_file',
]


@dataclass
class Codec(mixins.StreamValidationMixin, base.Dest, BaseWrapper):
    # noinspection PyUnresolvedReferences
    """
    Base class for output codecs.

    :arg codec: ffmpeg codec name.
    :arg bitrate: output bitrate in bps.
    """

    index = cast(int, base.Once('index'))
    """ Index of current codec in ffmpeg output streams."""

    codec: str = param(name='c', stream_suffix=True)
    bitrate: int = param(default=0, name='b', stream_suffix=True)

    def __post_init__(self) -> None:
        if self.codec is None:
            self.codec = self.__class__.codec
        super().__post_init__()

    @property
    def map(self) -> Optional[str]:
        """
        :returns: `-map` argument value depending of a node or a source
        connected to codec.
        """
        if self.edge is None:
            raise RuntimeError("Codec not connected to source")
        source = self.edge.input
        # Source input has name generated from input file index, stream
        # specifier and stream index. Node has no attribute index, so we use
        # Dest name ("[vout0]") as map argument. See also `Node.get_filter_args`
        return getattr(source, 'name', self.name)

    @property
    def connected(self) -> bool:
        """
        :return: True if codec is already connected to a node or a source.
        """
        return bool(self.edge)

    def get_args(self) -> List[bytes]:
        args = ['-map', self.map]
        return ensure_binary(args) + super().get_args()

    def clone(self, count: int = 1) -> List["Codec"]:
        """
        Creates multiple copies of self to reuse it as output node for multiple
        sources.

        Any connected input node is being split and connected to a corresponding
        copy of current filter.
        """
        if count == 1:
            return [self]
        raise RuntimeError("Trying to clone codec, is this intended?")

    def check_buffering(self) -> Optional[List[str]]:
        """
        Check that scenes read from input stream are ordered with ascending
        timestamps.

        :returns: A list of streams needed for this codec or None if metadata
            for codec can't be computed.
        """
        meta = self.get_meta_data()
        if not meta:
            return None
        prev = meta.scenes[0]
        for scene in meta.scenes[1:]:
            if prev.stream == scene.stream and prev.end > scene.start:
                # Previous scene in same stream is located after current, so
                # current decoded scene will be buffered until previous scene is
                # decoded.
                raise BufferError(prev, scene)
            prev = scene
        return meta.streams


@dataclass
class Output(BaseWrapper):
    # noinspection PyUnresolvedReferences
    """
    Base class for ffmpeg output.

    :arg codecs: list of codecs used in output.
    :arg format: output file format.
    :arg output_file: output file name.
    """
    codecs: List[Codec] = param(default=list, skip=True)
    no_video: Optional[bool] = param(name='vn')
    no_audio: Optional[bool] = param(name='an')
    format: str = param(name="f")
    output_file: str = param(name="", skip=True)

    def __lt__(self, other: base.InputType) -> "Output":
        """
        Connects a source or a filter to a first free codec.

        If there is no free codecs, new codec stub is created.
        """
        codec = self.get_free_codec(other.kind)
        other.connect_dest(codec)
        return self

    @property
    def video(self) -> Codec:
        """
        :returns: first video codec not connected to source.

        If no free codecs left, new one codec stub is appended to output.
        """
        return self.get_free_codec(VIDEO)

    @property
    def audio(self) -> Codec:
        """
        :returns: first audio codec not connected to source.

        If no free codecs left, new one codec stub is appended to output.
        """
        return self.get_free_codec(AUDIO)

    def get_free_codec(self, kind: StreamType, create: bool = True
                       ) -> Codec:
        """
        Finds first codec not connected to filter graph or to an input, or
        creates a new unnamed codec stub if no free codecs left.

        :param kind: desired codec stream type
        :param create: create new codec stub
        :return: first free codec or a new codec stub.
        """
        try:
            codec = next(filter(
                lambda c: not c.connected and c.kind == kind, self.codecs))
        except StopIteration:
            if not create:
                raise KeyError(kind)
            codec = Codec()
            codec.kind = kind
            self.codecs.append(codec)
        return codec

    def get_args(self) -> List[bytes]:
        """
        :returns: codec args and output file parameters for ffmpeg
        """
        args = []
        # Check if we need to disable audio or video for output file because no
        # corresponding codecs are found.
        # Skipping `-an` / `-vn` parameters is still supported by  manually
        # setting `no_audio` / `no_video` parameters to `False`.
        for codec in self.codecs:
            if codec.kind == VIDEO:
                self.no_video = False
            if codec.kind == AUDIO:
                self.no_audio = False
            args.extend(codec.get_args())
        if self.no_video is None:
            self.no_video = True
        if self.no_audio is None:
            self.no_audio = True
        args.extend(super().get_args())
        args.append(ensure_binary(self.output_file))
        return args


def output_file(filename: str, *codecs: Codec, **kwargs: Any) -> Output:
    """
    A shortcut to create proper output file.
    :param filename: output file name.
    :param codecs: codec list for this output.
    :param kwargs: output parameters.
    :return: configured ffmpeg output.
    """
    return Output(output_file=filename, codecs=list(codecs), **kwargs)


class OutputList(list):
    """ Supports unique output streams names generation."""

    def __init__(self, outputs: Iterable[Output] = ()) -> None:
        """
        :param outputs: list of output files
        """
        super().__init__()
        self.__video_index = 0
        self.__audio_index = 0
        self.extend(outputs)

    @property
    def codecs(self) -> List[Codec]:
        result: List[Codec] = []
        for output in self:
            result.extend(output.codecs)
        return result

    def append(self, output: Output) -> None:
        """
        Adds new output file to output list.

        :param output: output file
        """
        for codec in output.codecs:
            self.__set_index(codec)
        super().append(output)

    def extend(self, outputs: Iterable[Output]) -> None:
        """
        Adds multiple output files to output list.

        :param outputs: list of output files
        """
        for codec in chain(*map(lambda output: output.codecs, outputs)):
            self.__set_index(codec)
        super().extend(outputs)

    def get_args(self) -> List[bytes]:
        result: List[bytes] = []
        for source in self:
            result.extend(source.get_args())
        return result

    def __set_index(self, codec: Codec) -> None:
        if codec.kind == VIDEO:
            codec.index = self.__video_index
            self.__video_index += 1
        else:
            codec.index = self.__audio_index
            self.__audio_index += 1



from typing import TYPE_CHECKING, Optional

from fffw.graph import base, VIDEO

if TYPE_CHECKING:
    StreamValidationTarget = base.Dest
else:
    StreamValidationTarget = object


class StreamValidationMixin(StreamValidationTarget):
    hardware: Optional[str]

    def connect_edge(self, edge: base.Edge) -> base.Edge:
        self.validate_edge_kind(edge)
        self.validate_edge_device(edge)
        return super().connect_edge(edge)

    def validate_edge_kind(self, edge: base.Edge) -> None:
        kind = getattr(self, 'kind', None)
        if kind is None:
            return
        if edge.kind != kind:
            # Audio filter can't handle video stream and so on
            raise ValueError(edge.kind)

    def validate_edge_device(self, edge: base.Edge) -> None:
        if edge.kind != VIDEO:
            return
        meta = edge.get_meta_data(self)
        if meta is None:
            return
        try:
            filter_hardware = getattr(self, 'hardware')
        except AttributeError:
            # no hardware restrictions for filter/codec
            return
        device = getattr(meta, 'device', None)
        edge_hardware = None if device is None else device.hardware
        if filter_hardware != edge_hardware:
            # A stream uploaded to a video card could not be processed with CPU
            # filter.
            raise ValueError(edge_hardware)



import abc
from collections import Counter
from typing import Dict, Any, TypeVar, Type, overload
from typing import Optional, List, Union

from fffw.graph.meta import Meta, StreamType

InputType = Union["Source", "Node"]
OutputType = Union["Dest", "Node"]


class Traversable(metaclass=abc.ABCMeta):
    """
    Abstract class base for filter graph edges/nodes traversing and rendering.
    """

    @abc.abstractmethod
    def render(self, partial: bool = False) -> List[str]:
        """
        Returns a list of filter_graph edge descriptions.

        This method must be called in Namer context.

        :param partial: partially formatted graph render mode flag
        :return: edge description list ["[v:0]yadif[v:t1]", "[v:t1]scale[out]"]
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_meta_data(self, dst: OutputType) -> Optional[Meta]:
        """
        :param dst: destination node
        :return: metadata passed to destination node after transformation
        """
        raise NotImplementedError()


class Dest(Traversable):
    """
    Audio/video output stream node.

    Must connect to single filter output only.
    """
    kind: StreamType
    _edge: Optional["Edge"] = None

    def __repr__(self) -> str:
        return f"Dest('{self.name}')"

    @property
    def name(self) -> str:
        """
        :returns: edge name (i.e. [vout0]) for codec `-map` argument only.
        """
        if self._edge is None:
            raise RuntimeError("Dest not connected")
        return f'[{self._edge.name}]'

    @property
    def meta(self) -> Optional[Meta]:
        return self.get_meta_data(self)

    @property
    def edge(self) -> Optional["Edge"]:
        return self._edge

    def get_meta_data(self, dst: OutputType = None) -> Optional[Meta]:
        if self._edge is None:
            raise RuntimeError("Dest not connected")
        return self._edge.get_meta_data(self)

    def connect_edge(self, edge: "Edge") -> "Edge":
        """ Connects and edge to output stream.

        Should be called only from Node methods. Initializes edge identifier.

        :param edge: edge to connect output stream to.
        :type edge: Edge
        :return None
        """
        if not isinstance(edge, Edge):
            raise ValueError("Only edge allowed")
        if self._edge is not None:
            raise RuntimeError("Dest is already connected to %s" % self._edge)
        self._edge = edge
        return edge

    def render(self, partial: bool = False) -> List[str]:
        # Previous nodes/edges already rendered destination node.
        return []


class Edge(Traversable):
    """ Internal ffmpeg data stream graph."""

    # noinspection PyShadowingBuiltins
    def __init__(self, input: InputType, output: OutputType) -> None:
        """
        :param input: input node
        :param output: output node
        """
        super().__init__()
        self.__input = input
        self.__output = output

    def __repr__(self) -> str:
        return f'Edge#{self.name}[{self.input}, {self.output}]'

    @property
    def kind(self) -> StreamType:
        return self.__input.kind

    @property
    def input(self) -> InputType:
        return self.__input

    @property
    def output(self) -> OutputType:
        return self.__output

    @property
    def name(self) -> str:
        """
        Get actual name for edge from source node.

        Property must be accessed within Namer context.

        :returns: edge identifier generated from output node name if connected
        to Dest, or a  name of last enabled filter before (and including)
        current node.
        """
        if isinstance(self.output, Dest):
            # For final edges name is generated from destination node, like
            # [vout0] or [aout1]
            return Namer.name(self)
        # For edges connected to other filters disabled source nodes are skipped
        edge = self
        node = self.input
        while not getattr(node, 'enabled', True) and isinstance(node, Node):
            if node.inputs[0] is None:
                raise RuntimeError("Node input is None")
            edge = node.inputs[0]
            node = edge.input
        return Namer.name(edge)

    def _connect_source(self, src: InputType) -> None:
        """ Connects input node to the edge.

        :param src: source stream or filter output node.
        """
        if self.__input is not None:
            raise RuntimeError("edge already connected to input %s"
                               % self.__input)
        self.__input = src

    @overload
    def _connect_dest(self, dest: "Node") -> "Node":
        ...

    @overload
    def _connect_dest(self, dest: Dest) -> Dest:
        ...

    def _connect_dest(self, dest: OutputType) -> OutputType:
        """ Connects output node to the edge.

        :param dest: output stream or filter input node
        :return: connected node
        """
        if self.__output is not None:
            raise RuntimeError("edge already connected to output %s"
                               % self.__output)
        self.__output = dest
        return dest

    def get_meta_data(self, dst: OutputType) -> Optional[Meta]:
        return self.__input.get_meta_data(dst)

    def render(self, partial: bool = False) -> List[str]:
        if not self.__output:
            if partial:
                return []
            raise RuntimeError("output is none")
        return self.__output.render(partial=partial)

    def reconnect(self, dest: OutputType) -> None:
        """
        Allows to detach an edge from one output and connect to another one.
        """
        if isinstance(self.__output, Node):
            inputs = self.__output.inputs
            inputs[inputs.index(self)] = None
        self.__output = dest
        dest.connect_edge(self)


D = TypeVar('D', bound=Dest)


class Node(Traversable, abc.ABC):
    """ Graph node describing ffmpeg filter."""
    # Should be overridden in derived classes
    kind: StreamType  # filter type (VIDEO/AUDIO)
    filter: str  # filter name

    input_count: int = 1  # number of inputs
    output_count: int = 1  # number of outputs

    def __repr__(self) -> str:
        inputs = [f"[{str(i.name if i else '---')}]" for i in self.inputs]
        outputs = [f"[{str(i.name if i else '---')}]" for i in self.outputs]
        return f"{''.join(inputs)}{self.filter}{''.join(outputs)}"

    def __or__(self, other: "Node") -> "Node":
        """
        connect output edge to node
        :return: connected node
        """
        if not isinstance(other, Node):
            return NotImplemented
        return self.connect_dest(other)

    def __gt__(self, other: Dest) -> Dest:
        """
        connects output edge to destination
        :param other: destination codec
        :return: connected codec
        """
        if not isinstance(other, Dest):
            return NotImplemented
        return self.connect_dest(other)

    @property
    @abc.abstractmethod
    def args(self) -> str:
        raise NotImplementedError()

    @property
    def inputs(self) -> List[Optional[Edge]]:
        """
        :returns: list of placeholders for input edges.
        """
        if 'inputs' not in self.__dict__:
            self.__dict__['inputs'] = [None] * self.input_count
        return self.__dict__['inputs']

    @property
    def outputs(self) -> List[Optional[Edge]]:
        """
        :returns: list of placeholders for output edges.
        """
        if 'outputs' not in self.__dict__:
            self.__dict__['outputs'] = [None] * self.output_count
        return self.__dict__['outputs']

    @property
    def enabled(self) -> bool:
        return self.__dict__.get('enabled', True)

    @enabled.setter
    def enabled(self, value: bool) -> None:
        if not value:
            assert self.input_count == 1
            assert self.output_count == 1
        self.__dict__['enabled'] = value

    @property
    def meta(self) -> Optional[Meta]:
        """ Compute metadata for current node."""
        metadata = []
        for edge in self.inputs:
            if edge is None:
                raise RuntimeError("Input not connected")
            meta = edge.get_meta_data(self)
            if meta is None:
                return None
            metadata.append(meta)

        return self.transform(*metadata)

    # noinspection PyMethodMayBeStatic
    def transform(self, *metadata: Meta) -> Meta:
        """ Apply filter changes to stream metadata."""
        return metadata[0]

    def get_meta_data(self, dst: OutputType) -> Optional[Meta]:
        """ Returns metadata for selected destination."""
        for edge in self.outputs:
            if edge is None:
                continue
            if edge.output is dst:
                return self.meta
        else:
            raise KeyError(dst)

    def render(self, partial: bool = False) -> List[str]:
        if not self.enabled:
            # filter skipped, input is connected to output directly
            next_edge = self.outputs[0]
            if next_edge is None:
                if partial:
                    return []
                raise RuntimeError("output is None")
            return next_edge.render(partial=partial)

        result = [self.get_filter_cmd(partial=partial)]

        for dest in self.outputs:
            if dest is None:
                if partial:
                    continue
                raise RuntimeError("destination is none")
            if isinstance(dest.output, Dest):
                continue
            result.extend(dest.render(partial=partial))
        return result

    def get_filter_cmd(self, partial: bool = False) -> str:
        """
        Returns filter description.

        output format is like [IN] FILTER ARGS [OUT]
        where IN, OUT - lists of input/output edge id,
        FILTER - filter name, ARGS - filter params

        :param partial: partially formatted graph render mode flag
        :return: current description string like "[v:0]yadif[v:t1]"
        """
        inputs = []
        outputs = []
        for edge in self.inputs:
            if edge is None:
                raise RuntimeError("input is none")
            # Add Source name (i.e. 0:v) or unique node name (v:scale1) as
            # filter input.
            inputs.append(f'[{edge.name}]')

        for edge in self.outputs:
            if edge is None:
                if partial:
                    # outputs not connected, using a stub.
                    outputs.append('[---]')
                    continue
                raise RuntimeError("output is none")
            # Skip outgoing disabled nodes till Dest node, to get proper
            # output name if current node is last enabled before destination.
            node = edge.output
            while isinstance(node, Node) and not node.enabled:
                # if next node is disabled, use next edge id
                if node.outputs[0] is None and partial:
                    break
                edge = node.outputs[0]
                if edge is None:
                    raise RuntimeError("output is none")
                node = edge.output
            # Add unique output edge name (vout0 or a:volume1) to filter output
            outputs.append(f"[{edge.name}]")
        args = '=' + self.args if self.args else ''
        return ''.join(inputs) + self.filter + args + ''.join(outputs)

    def connect_edge(self, edge: "Edge") -> "Edge":
        """ Connects and edge to one of filter inputs

        :param edge: input stream edge
        :returns: connected edge
        """
        if not isinstance(edge, Edge):
            raise ValueError("only edge allowed")
        self.inputs[self.inputs.index(None)] = edge
        return edge

    @overload
    def connect_dest(self, other: "Node") -> "Node":
        ...

    @overload
    def connect_dest(self, other: Dest) -> Dest:
        ...

    def connect_dest(self, other: OutputType) -> OutputType:
        """ Connects next filter or output to one of filter outputs.

        :param other: next filter or output stream
        :return: next filter or output stream, connected to current stream
        """
        if not isinstance(other, (Node, Dest)):
            raise ValueError("only node and dest allowed")
        edge = Edge(input=self, output=other)
        self.outputs[self.outputs.index(None)] = edge
        other.connect_edge(edge)
        return other


N = TypeVar('N', bound=Node)


class Source(Traversable, metaclass=abc.ABCMeta):
    """ Graph node containing audio or video input.

    Must connect to single graph edge only as a source
    """

    def __init__(self, kind: StreamType,
                 meta: Optional[Meta] = None) -> None:
        """
        :param kind: stream type (VIDEO/AUDIO)
        :param meta: stream metadata
        """
        self._outputs: List[Edge] = []
        self._kind = kind
        self._meta = meta

    def __repr__(self) -> str:
        return f"Source('[{self.name}]')"

    def __or__(self, other: N) -> N:
        """
        Connect a filter to a source
        :return: connected filter
        """
        if not isinstance(other, Node):
            return NotImplemented
        return self.connect_dest(other)

    def __gt__(self, other: D) -> D:
        """
        Connects a codec to a source
        :param other: codec that will process current source stream
        :return: destination object
        """
        if not isinstance(other, Dest):
            return NotImplemented
        return self.connect_dest(other)

    @property
    def connected(self) -> bool:
        return bool(self._outputs)

    @property
    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError()

    @property
    def kind(self) -> StreamType:
        """
        :returns: stream type
        """
        return self._kind

    @property
    def meta(self) -> Optional[Meta]:
        """
        :returns: stream metadata
        """
        return self._meta

    @overload
    def connect_dest(self, other: N) -> N:
        ...

    @overload
    def connect_dest(self, other: D) -> D:
        ...

    def connect_dest(self, other: OutputType) -> OutputType:
        if not isinstance(other, (Node, Dest)):
            raise ValueError("only node or dest allowed")
        edge = Edge(input=self, output=other)
        other.connect_edge(edge)
        self._outputs.append(edge)
        return other

    def get_meta_data(self, dst: OutputType) -> Optional[Meta]:
        return self._meta

    def render(self, partial: bool = False) -> List[str]:
        result = []
        edge: Optional[Edge]
        for edge in self._outputs:
            node = edge.output
            # if output node is disabled, use next edge identifier.
            if isinstance(node, Node) and not node.enabled:
                edge = node.outputs[0]
                if edge is None:
                    if partial:
                        return []
                    raise RuntimeError("Skipped node is not ready for render")
            result.extend(edge.render(partial=partial))
        return result


Obj = TypeVar('Obj')


class Once:
    """ Property that must be set exactly once."""

    def __init__(self, attr_name: str) -> None:
        """
        :param attr_name: instance attribute name
        """
        self.attr_name = attr_name

    def __get__(self, instance: Obj, owner: Type[Obj]) -> Any:
        try:
            return instance.__dict__[self.attr_name]
        except KeyError:
            raise RuntimeError(f"{self.attr_name} is not initialized")

    def __set__(self, instance: Obj, value: Any) -> None:
        if self.attr_name in instance.__dict__:
            raise RuntimeError(f"{self.attr_name} already initialized")
        instance.__dict__[self.attr_name] = value


class Namer:
    """ Unique stream identifiers generator."""
    _stack: List["Namer"] = []

    @classmethod
    def name(cls, obj: Edge) -> str:
        current = cls._stack[0]
        return current._name(obj)

    def __init__(self) -> None:
        self._counters: Dict[str, int] = Counter()
        self._cache: Dict[int, str] = dict()

    def __enter__(self) -> "Namer":
        self._stack.append(self)
        return self._stack[0]

    def __exit__(self, *_: Any) -> None:
        self._stack.pop(-1)

    def _name(self, edge: Edge) -> str:
        """
        Generates name for an edge in filter graph.

        :param edge: edge that needs to be named
        :returns: unique Dest name if edge leads to destination (i.e. vout0),
        Source name if edge starts from input stream (i.e. 0:v) or unique
        input Node name generated from node filter (i.e. v:overlay1).
        """
        if id(edge) not in self._cache:
            src = edge.input
            dst = edge.output
            if isinstance(dst, Dest):
                prefix = f'{dst.kind.value}out'
                # generating unique edge id by dst kind
                name = f'{prefix}{self._counters[prefix]}'
                self._counters[prefix] += 1
            elif isinstance(src, Node):
                prefix = f'{src.kind.value}:{src.filter}'
                # generating unique edge id by src node kind and name
                name = f'{prefix}{self._counters[prefix]}'
                self._counters[prefix] += 1
            else:
                name = f'{src.name}'
            # caching edge name for idempotency
            self._cache[id(edge)] = name
        return self._cache[id(edge)]



from typing import *

from fffw.graph.meta import AUDIO, VIDEO, StreamType, Meta
from fffw.encoding import filters, inputs, outputs, FFMPEG

Group = Dict[int, Set[int]]
""" Grouping result."""
PairKey = Tuple[int, int]
""" Dict key containing ids of two related objects."""

Incoming = Union[inputs.Stream, filters.Filter]
Outgoing = Union[filters.Filter, outputs.Codec]


def group(first: Iterable[Any], second: Iterable[Any]) -> Group:
    """
    Group second iterable by first.

    :param first: First iterable. Id of each object in iterable will be used as
        a key.
    :param second: Second iterable. Id of each object in iterable will be added
        to a set, which is a value for corresponding key in first iterable.
    :returns: A dict that maps single object from first iterable to a set of
        objects from second iterable by object ids.
    """
    src_to_dst: Group = dict()
    for src, dst in zip(first, second):
        if src is None:
            continue
        dst_set = src_to_dst.setdefault(id(src), set())
        dst_set.add(id(dst))
    return src_to_dst


def prepare_src_splits(sources: Dict[int, Incoming],
                       groups: Group) -> Dict[PairKey, filters.Filter]:
    """
    Initialize split filter for each unique group.

    Source (which is a group key) is split to N nodes, where N is a number of
    corresponding outputs related to this source.

    :param sources: contains sources by object id.
    :param groups: destinations grouped by source id.
    :returns: split for each src/dst pair.
    """
    src_splits: Dict[PairKey, filters.Filter] = dict()
    for src_id, dst_set in groups.items():
        src = sources[src_id]
        # Split incoming stream for each destination that will be connected
        # to it.
        for dst_id, s in zip(dst_set, src.split(len(dst_set))):
            src_splits[src_id, dst_id] = s
    return src_splits


def prepare_dst_clones(destinations: Dict[int, Outgoing],
                       groups: Group) -> Dict[PairKey, Outgoing]:
    """
    Prepares copies of filter for each unique group.

    Destination (which is a group key) is copied N times with same parameters,
    where N is a number of unique inputs related to this destination.

    If destination is a filter with multiple inputs, each input is split
    multiple times, disconnected from initial filter and split results are
    connected to filter copies.

    :param destinations: contains outputs by object id.
    :param groups: sources grouped by destination id.
    :returns: destination filter copy for each src/dst pair.
    """
    dst_clones: Dict[Tuple[int, int], Outgoing] = dict()
    for dst_id, src_set in groups.items():
        dst = destinations[dst_id]
        # Clone destination filter for each source that will be connected
        # to it.
        # noinspection PyTypeChecker
        clones = cast(List[Outgoing], dst.clone(len(src_set)))
        for src_id, c in zip(src_set, clones):
            dst_clones[dst_id, src_id] = c
    return dst_clones


def map_sources_to_destinations(
        sources: List[Incoming],
        src_splits: Dict[Tuple[int, int], filters.Filter],
        destinations: List[Optional[Outgoing]],
        dst_clones: Dict[Tuple[int, int], Outgoing]
) -> "Vector":
    """
    Links sources to destinations.

    For each source/destination pair we find corresponding source split result
    and destination clone. Split is connected to clone, clone is added to
    resulting vector and cached to prevent connecting same split to same clone
    twice.

    If destination is None, which means that it is disabled by mask,
    split is added to resulting vector instead of clone.

    :param sources: list of initial incoming nodes.
    :param src_splits: contains actual split node for each source/destination
        pair.
    :param destinations: list of outgoing nodes (or None if disabled by mask).
    :param dst_clones: contains actual dst clone for each source/destination
        pair.
    :returns: final vector with results of src/dst connection.
    """
    links: Dict[Tuple[int, int], Outgoing] = dict()
    results: List[Outgoing] = list()
    # connecting sources to destinations and gathering results
    for src, dst in zip(sources, destinations):
        # use split instead of initial incoming node
        split = src_splits[id(src), id(dst)]
        if dst is None:
            # Skip via mask, pass split to output. We can't use src here
            # because it is already connected to split filter
            results.append(split)
            continue
        clone = dst_clones[id(dst), id(src)]

        key = id(split), id(clone)

        if key not in links:
            # connect same src to same dst only once
            # noinspection PyTypeChecker
            links[key] = cast(Outgoing, split.connect_dest(clone))

        # add destination node to results
        results.append(links[key])
    # noinspection PyTypeChecker
    return Vector(cast(Union[List[filters.Filter], List[outputs.Codec]],
                       results))


def init_filter_vector(filter_class: Type[filters.Filter],
                       params: Union[
                           List[Dict[str, Any]],
                           List[List[Any]],
                           List[Any],
                       ]) -> "Vector":
    """
    Initializes filter vector from filter class and list of parameters.

    :param filter_class: type to initialize
    :param params: list of filter class positional or keyword arguments
    :returns: filter vector with elements count equal to length of params list,
        each vector is instance of filter_class.
    """
    vector = []
    # noinspection PyTypeChecker
    factory = cast(Callable[..., filters.Filter], filter_class)
    seen_filters: Dict[str, filters.Filter] = dict()
    for param in params:
        try:
            f = seen_filters[repr(param)]
        except KeyError:
            if isinstance(param, dict):
                f = factory(**param)
            elif hasattr(param, '__iter__'):
                f = factory(*param)
            else:
                f = factory(param)
            seen_filters[repr(param)] = f
        vector.append(f)
    return Vector(vector)


def normalize_args(dst: Union[filters.Filter,
                              Type[filters.Filter],
                              "Vector"],
                   mask: Optional[List[bool]] = None,
                   params: Union[
                       None,
                       List[Dict[str, Any]],
                       List[List[Any]],
                       List[Any],
                   ] = None) -> Tuple["Vector", List[bool]]:
    """
    Transform different args to same form: vector of filters/codecs and
    boolean mask with corresponding length.

    :param dst: destination vector, single filter or filter class used to
        initialize same filter with different parameters.
    :param mask: list of flags that allows to skip applying destination
        filters to some of sources.
    :param params: filter class constructor arguments used to initialize
        a destination filter vector from filter class.
    :returns: new vector which is a result of applying destination filters
        to a input streams vector.
    """
    if isinstance(dst, type):
        # handle filter class + params
        assert issubclass(dst, filters.Filter), "filter class needed"
        assert params is not None, "params not specified for filter class"
        dst = init_filter_vector(dst, params)
    elif isinstance(dst, filters.Filter):
        # handle filter instance
        dst = Vector(dst)
    elif not isinstance(dst, Vector):
        # handle vector instance - rest types are not supported.
        raise TypeError(dst)

    # align mask and destination lengths
    if mask is not None:
        if len(dst) == 1:
            # handle filter instance and mask vector
            dst = Vector(dst * len(mask))
        assert len(dst) == len(mask)
    else:
        # handle omitted mask
        mask = [True] * len(dst)
    return dst, mask


class Vector(tuple):
    """
    Represents immutable stream vector that helps to apply same or similar
    filters to a set of stream simultaneously.
    """

    def __new__(cls, source: Union[inputs.Stream, filters.Filter,
                                   Iterable[filters.Filter],
                                   Iterable[outputs.Codec]]) -> "Vector":
        """
        Constructs new Vector from single input or a list of inputs.

        Normalizes input parameter to an iterable of input stream, filter or
        codec.

        :param source: single input stream (video or audio), single filter,
            list of filters or list of codecs.
        :returns: an immutable list of streams, filters or codecs.
        """
        iterable: Union[Iterable[inputs.Stream],
                        Iterable[filters.Filter],
                        Iterable[outputs.Codec]]
        if isinstance(source, filters.Filter):
            iterable = [source]
        elif isinstance(source, inputs.Stream):
            # This branch is separated from previous for mypy.
            iterable = [source]
        else:
            iterable = source
        return tuple.__new__(cls, iterable)  # noqa

    def __or__(self, other: Union["Vector", filters.Filter]) -> "Vector":
        """ A shortcut to connect vector to another filter."""
        return self.connect(other)

    def __ror__(self, other: filters.Filter) -> "Vector":
        # noinspection PyUnresolvedReferences
        """ A shortcut to connect a filter to the vector.

        >>> overlay: Vector = simd | Overlay(1100, 100)
        >>> scaled_logo: Filter = logo.video | Scale(120, 120)
        >>> scaled_logo | overlay
        """
        if not isinstance(other, filters.Filter):
            return NotImplemented
        return Vector(other) | self

    @property
    def kind(self) -> StreamType:
        """
        :returns: a kind of streams in vector.
        """
        kinds = {s.kind for s in self}
        if len(kinds) != 1:
            raise RuntimeError("multiple kind of streams in vector")
        return self[0].kind

    @property
    def meta(self) -> Meta:
        """
        :returns: metadata for a stream in vector.
        """
        if len(self) != 1:
            raise RuntimeError("not a scalar")
        return self[0].meta

    @overload
    def connect(self, dst: filters.Filter, mask: Optional[List[bool]] = None
                ) -> "Vector":
        """

        >>> vector = Vector(inputs.Stream(VIDEO))
        >>> vector.connect(filters.Scale(), mask=[True, False])
        """
        ...  # pragma: no cover

    @overload
    def connect(self, dst: Type[filters.Filter],
                mask: Optional[List[bool]] = None,
                *,
                params: Union[List[Dict[str, Any]], List[List[Any]], List[Any]]
                ) -> "Vector":
        """

        >>> vector = Vector(inputs.Stream(VIDEO))
        >>> vector.connect(filters.Scale, params=[
        ... {'width': 1280, 'height': 720},
        ... {'width': 640, 'height': 360}])
        >>>
        """
        ...  # pragma: no cover

    @overload
    def connect(self, dst: "Vector", mask: Optional[List[bool]] = None
                ) -> "Vector":
        """

        >>> vector = Vector(inputs.Stream(VIDEO))
        >>> out = Vector([
        ... outputs.Codec(codec='libx264'),
        ... outputs.Codec(codec='libx264')])
        >>> vector.connect(out)
        """
        ...  # pragma: no cover

    def connect(self, dst: Union[filters.Filter,
                                 Type[filters.Filter],
                                 "Vector"],
                mask: Optional[List[bool]] = None,
                params: Union[
                    None,
                    List[Dict[str, Any]],
                    List[List[Any]],
                    List[Any],
                ] = None) -> "Vector":
        """
        Connects current vector to destination filter vector.

        :param dst: destination vector, single filter or filter class used to
            initialize same filter with different parameters.
        :param mask: list of flags that allows to skip applying destination
            filters to some of sources.
        :param params: filter class constructor arguments used to initialize
            a destination filter vector from filter class.
        :returns: new vector which is a result of applying destination filters
            to a input streams vector.
        """
        # Transform different args to same form: vector of filters/codecs and
        # boolean mask with corresponding length.
        dst, mask = normalize_args(dst, mask, params)

        # input list
        sources: List[Union[Incoming]] = list(self)
        # filter list
        destinations: List[Optional[Outgoing]] = list(dst)

        if len(sources) != len(destinations):
            # Transform source or destination vector to fit vector length.
            # We support connecting single source to multiple destinations or
            # multiple sources to a single destination.
            if len(sources) == 1:
                # adjusting single source to multiple destinations
                sources = sources * len(destinations)
            elif len(destinations) == 1:
                # adjusting single destination to multiple sources
                destinations = destinations * len(sources)
            else:
                raise RuntimeError("Can't apply M sources to N destinations")

        # disabling destinations by mask
        for i, enabled in enumerate(mask):
            if not enabled:
                destinations[i] = None

        # sources cache
        src_by_id: Dict[int, Incoming] = {id(src): src for src in sources}
        # destinations cache
        dst_by_id: Dict[int, Outgoing] = {
            id(dst): dst for dst in destinations if dst is not None}

        # Group destinations by source to prepare Split() filters at next step.
        src_to_dst = group(sources, destinations)

        # Group sources by destination to prepare destination clone() calls.
        dst_to_src = group(destinations, sources)

        # Split sources to have unique input for each unique destination
        # connected to same source
        src_splits = prepare_src_splits(src_by_id, src_to_dst)

        # Split destinations (with their inputs) to have unique output for each
        # unique source connected to same destination
        dst_clones = prepare_dst_clones(dst_by_id, dst_to_src)

        # Link split sources to cloned destinations.
        return map_sources_to_destinations(sources, src_splits,
                                           destinations, dst_clones)


class FFMPEGFactory:
    """ MyPy-respected FFMPEG factory definition."""

    def __call__(self, **kwargs: Any) -> FFMPEG:
        return FFMPEG(**kwargs)


class SIMD:
    """
    Single Instruction Multiple Data helper for video file processing.

    Handles Vector initialization from source streams and output codecs
    connections.
    """
    ffmpeg_wrapper = FFMPEGFactory()

    def __init__(self, source: inputs.Input, *results: outputs.Output,
                 **kwargs: Any) -> None:
        """
        :param source: input file for ffmpeg.
        :param results: list of ffmpeg output files.
        :param kwargs: ffmpeg command line arguments.
        """
        self.validate_input_file(source)
        for output in results:
            self.validate_output_file(output)
        self.__source = source
        self.__extra: List[inputs.Input] = []
        self.__results = results
        self.__ffmpeg: Optional[FFMPEG] = None
        self.__kwargs = kwargs

    @overload
    def __lt__(self, other: Union[inputs.Stream, filters.Filter, Vector]
               ) -> Vector:
        ...

    @overload
    def __lt__(self, other: inputs.Input) -> inputs.Input:
        ...

    def __lt__(self, other: Union[Vector, inputs.Input, inputs.Stream,
                                  filters.Filter]
               ) -> Union[Vector, inputs.Input]:
        # noinspection PyUnresolvedReferences
        """
        A shortcut to connect additional input file or codec vector.

        >>> simd = SIMD(inputs.input_file('input.mp4'))
        >>> # Adding extra input file
        >>> logo = simd < inputs.input_file('logo.png')
        >>> # Finalizing filter to simd with single stream
        >>> simd | filters.Scale(1280, 720) > simd
        >>> # Finalizing input stream excluded from filter graph
        >>> preroll.audio > simd
        >>> # Finalizing stream vector
        >>> scaled_vector = simd.video.connect(Scale, params=[size1, size2])
        >>> scaled_vector > simd
        """
        if isinstance(other, (inputs.Stream, filters.Filter)):
            # finalizing stream excluded from filter graph or single filtered
            # stream
            other = Vector(other)
        if isinstance(other, Vector):
            return other.connect(self.get_codecs(other.kind))
        elif isinstance(other, inputs.Input):
            return self.add_input(other)
        else:
            # noinspection PyTypeChecker
            return NotImplemented

    def __or__(self, other: filters.Filter) -> Vector:
        """
        A shortcut to connect input file stream to a filter.

        >>> simd = SIMD(inputs.input_file('input.mp4'))
        >>> vector = simd | filters.Scale(1280, 720)
        """
        if not isinstance(other, filters.Filter):
            return NotImplemented
        return Vector(self.get_stream(other.kind)).connect(other)

    @staticmethod
    def validate_input_file(input_file: inputs.Input) -> None:
        """
        Checks that input file contains streams information with stream
        metadata.
        """
        if not input_file.streams:
            raise ValueError("streams must be set for input file")
        for stream in input_file.streams:
            if not stream.meta:
                raise ValueError("stream metadata must be set for input file")

    @staticmethod
    def validate_output_file(output: outputs.Output) -> None:
        """
        Checks that output file contains codec information.
        """
        if not output.codecs:
            raise ValueError("codecs must be set for output file")

    @property
    def ffmpeg(self) -> FFMPEG:
        """
        Initializes ffmpeg wrapper with input/output files and command line
        arguments.

        Note that ffmpeg instance can't be instantiated twice, because existing
        output files could be connected to input streams only once.

        :returns: cached FFMPEG instance.
        """
        if self.__ffmpeg is None:
            self.__ffmpeg = self.ffmpeg_wrapper(
                input=self.__source, **self.__kwargs)

            for source in self.__extra:
                self.__ffmpeg.add_input(source)

            for output in self.__results:
                for codec in output.codecs:
                    if codec.connected:
                        continue
                    self.get_stream(codec.kind).connect_dest(codec)
                self.__ffmpeg.add_output(output)

        return self.__ffmpeg

    @property
    def video(self) -> Vector:
        """
        :returns: a vector with single video input stream
        """
        return Vector(self.get_stream(VIDEO))

    @property
    def audio(self) -> Vector:
        """
        :returns: a vector with single audio input stream
        """
        return Vector(self.get_stream(AUDIO))

    def get_stream(self, kind: StreamType) -> inputs.Stream:
        """
        :param kind: desired stream kind
        :return: first stream of desired kind from input file
        :raises KeyError: if no streams of this kind found.
        """
        for stream in self.__source.streams:
            if stream.kind == kind:
                return stream
        raise KeyError(kind)

    def get_codecs(self, kind: StreamType) -> Vector:
        """
        :param kind: desired vector kind
        :return: a vector of all codecs of desired kind for each output.
        """
        result = []
        for output in self.__results:
            result.append(output.get_free_codec(kind, create=False))
        return Vector(result)

    def add_input(self, source: inputs.Input) -> inputs.Input:
        """
        Adds additional input file to ffmpeg
        :param source: additional input file
        :returns: connected input
        """
        self.validate_input_file(source)
        self.__extra.append(source)
        return source



from dataclasses import dataclass
from typing import List, Optional, Union

from fffw.encoding import inputs
from fffw.encoding.complex import FilterComplex
from fffw.encoding.inputs import InputList, Input
from fffw.encoding.outputs import OutputList, Output, Codec
from fffw.graph import base, meta
from fffw.graph.meta import AUDIO, VIDEO, StreamType
from fffw.wrapper import BaseWrapper, ensure_binary, param

__all__ = ['FFMPEG']


@dataclass
class FFMPEG(BaseWrapper):
    """
    ffmpeg command line basic wrapper.

    >>> from fffw.encoding.codecs import VideoCodec, AudioCodec
    >>> from fffw.encoding.filters import Scale
    >>> from fffw.encoding.outputs import output_file
    >>> ff = FFMPEG('/tmp/input.mp4', overwrite=True)
    >>> c = VideoCodec('libx264', bitrate=4_000_000)
    >>> ff.video | Scale(1280, 720) > c
    VideoCodec(codec='libx264', bitrate=4000000)
    >>> ff.overwrite = True
    >>> ff > output_file('/tmp/output.mp4', c,
    ...                  AudioCodec('libfdk_aac', bitrate=192_000))
    >>> ff.get_cmd()
    'ffmpeg -y -i /tmp/input.mp4\
 -filter_complex "[0:v]scale=w=1280:h=720[vout0]"\
 -map "[vout0]" -c:v libx264 -b:v 4000000 -map 0:a -c:a libfdk_aac -b:a 192000\
 /tmp/output.mp4'
    >>>
    """
    command = 'ffmpeg'
    stderr_markers = ['[error]', '[fatal]']
    input: Union[str, Input] = param(skip=True)
    output: Union[str, Output] = param(skip=True)

    loglevel: str = param()
    """ Loglevel: i.e. `level+info`."""
    overwrite: bool = param(name='y')
    """ Overwrite output files without manual confirmation."""
    init_hardware: str = param(name='init_hw_device')
    """ Initializes hardware acceleration device."""
    filter_hardware: str = param(name='filter_hw_device')
    """ Sets a device for filter graph by it's name set with `init_hardware`."""

    def __post_init__(self) -> None:
        """
        Fills internal shared structures for input and output files, and
        initializes filter graph.
        """

        self.__inputs = InputList()
        if self.input:
            if not isinstance(self.input, Input):
                self.__inputs.append(Input(input_file=self.input))
            else:
                self.__inputs.append(self.input)

        self.__outputs = OutputList()
        if self.output:
            if not isinstance(self.output, Output):
                self.__outputs.append(Output(output_file=self.output))
            else:
                self.__outputs.append(self.output)

        self.__filter_complex = FilterComplex(self.__inputs, self.__outputs)

        # calling super() to freeze params.
        super().__post_init__()

    def __lt__(self, other: Input) -> Input:
        """ Adds new source file.

        >>> ff = FFMPEG()
        >>> src = ff < Input(input_file='/tmp/input.mp4')
        >>>
        """
        if not isinstance(other, Input):
            return NotImplemented
        return self.add_input(other)

    def __gt__(self, other: Output) -> Output:
        """ Adds new output file.

        >>> from fffw.encoding.inputs import *
        >>> from fffw.encoding.outputs import *
        >>> ff = FFMPEG(input=input_file('input.mp4'))
        >>> dest = ff > output_file('/tmp/output.mp4')
        >>>
        """
        if not isinstance(other, Output):
            return NotImplemented
        return self.add_output(other)

    @property
    def filter_device(self) -> meta.Device:
        """ Returns filter hardware device metadata."""
        hardware, init = self.init_hardware.split("=")
        name = init.split(':', 1)[0]
        if self.filter_hardware != name:
            raise ValueError(self.filter_hardware)
        return meta.Device(hardware, name)

    @property
    def video(self) -> inputs.Stream:
        """
        :returns: first video stream not yet connected to filter graph or codec.

        >>> from fffw.encoding.filters import Scale
        >>> ff = FFMPEG('/tmp/input.mp4')
        >>> ff.video | Scale(1280, 720)
        Scale(width=1280, height=720)
        >>>
        """
        return self._get_free_source(VIDEO)

    @property
    def audio(self) -> inputs.Stream:
        """

        :returns: first audio stream not yet connected to filter graph or codec.

        >>> from fffw.encoding.codecs import AudioCodec
        >>> ff = FFMPEG('/tmp/input.mp4')
        >>> ac = AudioCodec('aac')
        >>> ff.audio > ac
        AudioCodec(codec='aac', bitrate=0)
        >>>
        """
        return self._get_free_source(AUDIO)

    def _get_free_source(self, kind: StreamType) -> inputs.Stream:
        """
        :param kind: stream type
        :return: first stream of this kind not connected to destination
        """
        for stream in self.__inputs.streams:
            if stream.kind != kind or stream.connected:
                continue
            return stream
        else:
            raise RuntimeError("no free streams")

    def _add_codec(self, c: Codec) -> Optional[Codec]:
        """ Connect codec to filter graph output or input stream.

        :param c: codec to connect to free source
        :returns: None of codec already connected to filter graph or codec
            itself if it was connected successfully to input stream.
        """
        if c.connected:
            return None
        node = self._get_free_source(c.kind)
        node.connect_dest(c)
        return c

    def get_args(self) -> List[bytes]:
        """
        :returns: command line arguments for ffmpeg.

        This includes:
        - ffmpeg executable name
        - ffmpeg parameters
        - input list args
        - filter_graph definition
        - output list args
        """

        with base.Namer():
            fc = str(self.__filter_complex)
            fc_args = ['-filter_complex', fc] if fc else []

            # Namer context is used to generate unique output stream names
            return (super().get_args() +
                    self.__inputs.get_args() +
                    ensure_binary(fc_args) +
                    self.__outputs.get_args())

    def add_input(self, input_file: Input) -> Input:
        """ Adds new source to ffmpeg.

        >>> ff = FFMPEG()
        >>> ff.add_input(Input(input_file="/tmp/input.mp4"))
        >>>
        """
        assert isinstance(input_file, Input)
        self.__inputs.append(input_file)
        return input_file

    def add_output(self, output: Output) -> Output:
        """
        Adds output file to ffmpeg and connect it's codecs to free sources.

        >>> ff = FFMPEG()
        >>> ff.add_output(Output(output_file='/tmp/output.mp4'))
        >>>
        """
        self.__outputs.append(output)
        for codec in output.codecs:
            self._add_codec(codec)
        return output

    def handle_stderr(self, line: str) -> str:
        """
        Handle ffmpeg output.

        Capture only lines containing one of `stderr_markers`.

        :param line: ffmpeg output line
        :returns: line to be appended to whole ffmpeg output.
        """
        if not self.stderr_markers:
            # if no markers are defined, handle each line
            return super().handle_stderr(line)
        # capture only lines containing markers
        for marker in self.stderr_markers:
            if marker in line:
                return super().handle_stderr(line)
        return ''

    def check_buffering(self) -> None:
        """
        Checks that ffmpeg command will not cause frame buffering and
        out of memory errors.

        Each input file must be read simultaneously be all codecs in outputs,
        or some streams will be buffered until requested by output codecs.
        """
        chains = []
        for output in self.__outputs:
            for codec in output.codecs:
                streams = codec.check_buffering()
                if streams is None:
                    # Streams can't be computed because of missing metadata.
                    raise ValueError(streams)
                chains.append(streams)
        for chunk in zip(*chains):
            # Check that every codec reads same input stream
            if len(set(chunk)) > 1:
                # some codec read different file at this step, so one of input
                # stream will be buffered until this file is read by another
                # codec.
                raise BufferError(chunk)
